@@ -7,6 +7,8 @@ from multiprocessing import Pool
 import functools
 import re
 import gc
+import tempfile
+import os
 
 
 def writeFragments(fragments, filepath):
@@ -19,12 +21,10 @@ def writeFragments(fragments, filepath):
     filepath : str
         Path for output file
     """
-    with open(filepath, "w") as outf:
+    with open(filepath, "a") as outf:
         for i in fragments:
-            for j in i:
-                for y in j:
-                    outstr = "\t".join(map(str, y))
-                    outf.write(outstr + "\n")
+            outstr = "\t".join(map(str, i))
+            outf.write(outstr + "\n")
 
 
 def createPositionLookup(frags, pos=1):
@@ -75,13 +75,13 @@ def collapseOverlapFragments(counts, pos=1):
 def collapseFragments(fragments):
     """Collapse duplicate fragments
     """
-    fraglist = [list(x.values()) for x in list(fragments.values())]
+    fraglist = list(fragments.values())
     fragcoords_with_bc = ["|".join(map(str, x)) for x in fraglist]
     counts = Counter(fragcoords_with_bc)
-    
+
     if len(fraglist) == 0:
         return list()
-    
+
     # enumerate fragments and barcodes
     frag_id_lookup = id_lookup(l=["|".join(map(str, x[:3])) for x in fraglist])
     bc_id_lookup = id_lookup(l=[x[3] for x in fraglist])
@@ -100,7 +100,7 @@ def collapseFragments(fragments):
         bcstr = i[0].split("|")[3]
         row.append(frag_id_lookup["|".join(rowstr)])
         col.append(bc_id_lookup[bcstr])
-    
+
     # free memory
     del fraglist
     del fragments
@@ -144,7 +144,14 @@ def id_lookup(l):
 
 
 def getFragments(
-    interval, bam, min_mapq=30, cellbarcode="CB", readname_barcode=None, cells=None, max_distance=5000
+    interval,
+    bam,
+    min_mapq=30,
+    cellbarcode="CB",
+    readname_barcode=None,
+    cells=None,
+    max_distance=5000,
+    chunksize=500000,
 ):
     """Extract ATAC fragments from BAM file
 
@@ -167,9 +174,16 @@ def getFragments(
         Maximum distance between integration sites for the fragment to be retained.
         Allows filtering of implausible fragments that likely result from incorrect 
         mapping positions. Default is 5000 bp.
+    chunksize : int
+        Number of BAM entries to read through before collapsing and writing
+        fragments to disk. Higher chunksize will use more memory but will be 
+        faster.
     """
     fragment_dict = dict()
     inputBam = pysam.AlignmentFile(bam, "rb")
+    outfile = tempfile.NamedTemporaryFile(delete=False)
+    outname = outfile.name
+    x = 0
     if readname_barcode is not None:
         readname_barcode = re.compile(readname_barcode)
     for i in inputBam.fetch(interval[0], 0, interval[1]):
@@ -180,15 +194,81 @@ def getFragments(
             cellbarcode=cellbarcode,
             readname_barcode=readname_barcode,
             cells=cells,
+            max_dist=max_distance,
         )
-    fragment_dict = filterFragmentDict(fragments=fragment_dict, max_distance=max_distance)
-    collapsed = collapseFragments(fragments=fragment_dict)
-    gc.collect()
-    return collapsed
+        x += 1
+        if x > chunksize:
+            current_position = i.reference_start
+            complete = findCompleteFragments(
+                fragments=fragment_dict,
+                max_dist=max_distance,
+                current_position=current_position,
+                max_collapse_dist=20,
+            )
+            collapsed = collapseFragments(fragments=complete)
+            writeFragments(fragments=collapsed, filepath=outname)
+            x = 0
+            gc.collect()
+    # collapse and write the remaining fragments
+    complete = findCompleteFragments(
+        fragments=fragment_dict,
+        max_dist=max_distance,
+        current_position=i.reference_start,
+        max_collapse_dist=-max_distance,  # make sure we get them all
+    )
+    collapsed = collapseFragments(fragments=complete)
+    writeFragments(fragments=collapsed, filepath=outname)
+    return outname
+
+
+def findCompleteFragments(fragments, max_dist, current_position, max_collapse_dist=20):
+    """Find complete fragments that are >max_dist bp away from
+    the current BAM file position
+    
+    Parameters
+    ----------
+    fragments : dict
+        A dictionary containing ATAC fragment information
+    max_dist : int
+        The maximum allowed distance between fragment start and 
+        end positions
+    current_position : int
+        The current position being looked at in the position-sorted
+        BAM file
+    max_collapse_dist : int
+        Maximum allowed distance for fragments from the same cell 
+        barcode that share one Tn5 integration site (fragment 
+        start or end coordinate) to be collapsed into a single 
+        fragment.
+    
+    Moves completed fragments to a new dictionary
+    Completed fragments will be deleted from the original dictionary
+    """
+    allkey = list(fragments.keys())
+    completed = dict()
+    d = max_dist + max_collapse_dist
+    for key in allkey:
+        if fragments[key][4]:  # complete fragment
+            if (fragments[key][2] + d) < current_position:
+                completed[key] = fragments[key][:-1]  # removes "completed" T/F information
+                del fragments[key]
+        else:
+            # remove incomplete fragments that are
+            # too far away to ever be complete
+            if fragments[key][1] is None:
+                if (fragments[key][2] + d) < current_position:
+                    del fragments[key]
+            elif fragments[key][2] is None:
+                if (fragments[key][1] + d) < current_position:
+                    del fragments[key]
+            else:
+                # start and end coordinates present without a cell barcode
+                del fragments[key]
+    return completed
 
 
 def updateFragmentDict(
-    fragments, segment, min_mapq, cellbarcode, readname_barcode, cells
+    fragments, segment, min_mapq, cellbarcode, readname_barcode, cells, max_dist
 ):
     """Update dictionary of ATAC fragments
     Takes a new aligned segment and adds information to the dictionary,
@@ -213,6 +293,8 @@ def updateFragmentDict(
         use the read tags.
     cells : list
         List of cells to retain. If None, retain all cells found.
+    max_dist : int
+        Maximum allowed distance between fragment start and end sites
     """
     # because the cell barcode is not stored with each read pair (only one of the pair)
     # we need to look for each read separately rather than using the mate cigar / mate postion information
@@ -233,7 +315,7 @@ def updateFragmentDict(
     rend = segment.reference_end
     qstart = segment.query_alignment_start
     is_reverse = segment.is_reverse
-    if rend is None:
+    if (rend is None) or (rstart is None):
         return fragments
     # correct for soft clipping
     rstart = rstart + qstart
@@ -242,34 +324,80 @@ def updateFragmentDict(
         rend = rend - 5
     else:
         rstart = rstart + 4
-    if qname in fragments.keys():
-        if is_reverse:
-            fragments[qname]["end"] = rend
-        else:
-            fragments[qname]["start"] = rstart
-    else:
-        fragments[qname] = {
-            "chrom": chromosome,
-            "start": None,
-            "end": None,
-            "cell": cell_barcode,
-        }
-        if is_reverse:
-            fragments[qname]["end"] = rend
-        else:
-            fragments[qname]["start"] = rstart
+    fragments = addToFragments(
+        fragments, qname, chromosome, rstart, rend, cell_barcode, is_reverse, max_dist
+    )
     return fragments
 
 
-def filterFragmentDict(fragments, max_distance=5000):
-    """Remove invalid entries"""
-    allkey = list(fragments.keys())
-    for key in allkey:
-        vals = list(fragments[key].values())
-        if not all(vals):
-            del fragments[key]
-        elif (vals[2] - vals[1]) > max_distance:
-            del fragments[key]
+def addToFragments(
+    fragments, qname, chromosome, rstart, rend, cell_barcode, is_reverse, max_dist
+):
+    """Add new fragment information to dictionary
+    
+    Parameters
+    ----------
+
+    fragments : dict
+        A dictionary containing all the fragment information
+    qname : str
+        Read name
+    chromosome : str
+        Chromosome name
+    rstart : int
+        Alignment start position
+    rend : int
+        Alignment end position
+    cell_barcode : str
+        Cell barcode sequence
+    is_reverse : bool
+        Read is aligned to reverse strand
+    max_dist : int
+        Maximum allowed fragment size
+    """
+    if qname in fragments.keys():
+        if is_reverse:
+            current_coord = fragments[qname][1]
+            if current_coord is None:
+                # read aligned to the wrong strand, don't include
+                del fragments[qname]
+            elif ((rend - current_coord) > max_dist) or ((rend - current_coord) < 0):
+                # too far away, don't include
+                del fragments[qname]
+            else:
+                if cell_barcode is None and fragments[qname][3] is None:
+                    # both fragment ends present but no cell barcode
+                    del fragments[qname]
+                else:
+                    fragments[qname][2] = rend
+                    fragments[qname][4] = True
+        else:
+            current_coord = fragments[qname][2]
+            if current_coord is None:
+                del fragments[qname]
+            elif ((current_coord - rstart) > max_dist) or (
+                (current_coord - rstart) < 0
+            ):
+                del fragments[qname]
+            else:
+                if cell_barcode is None and fragments[qname][3] is None:
+                    del fragments[qname]
+                else:
+                    fragments[qname][1] = rstart
+                    fragments[qname][4] = True
+    else:
+        # new read pair, add to dictionary
+        fragments[qname] = [
+            chromosome,     # chromosome    0
+            None,           # start         1
+            None,           # end           2
+            cell_barcode,   # cell          3
+            False,          # complete      4
+        ]
+        if is_reverse:
+            fragments[qname][2] = rend
+        else:
+            fragments[qname][1] = rstart
     return fragments
 
 
@@ -283,6 +411,7 @@ def fragments(
     readname_barcode=None,
     cells=None,
     max_distance=5000,
+    chunksize=500000,
 ):
     """Create ATAC fragment file from BAM file
 
@@ -316,6 +445,10 @@ def fragments(
         Maximum distance between integration sites for the fragment to be retained.
         Allows filtering of implausible fragments that likely result from incorrect 
         mapping positions. Default is 5000 bp.
+    chunksize : int
+        Number of BAM entries to read through before collapsing and writing
+        fragments to disk. Higher chunksize will use more memory but will be 
+        faster.
     """
     nproc = int(nproc)
     chrom = utils.get_chromosomes(bam, keep_contigs=chromosomes)
@@ -330,9 +463,18 @@ def fragments(
                 cellbarcode=cellbarcode,
                 readname_barcode=readname_barcode,
                 cells=cells,
-                max_distance=max_distance
+                max_distance=max_distance,
+                chunksize=chunksize,
             ),
             list(chrom.items()),
         )
     ]
-    writeFragments(fragments=[res.get() for res in frag_lists], filepath=fragment_path)
+    filenames = [res.get() for res in frag_lists]
+    # cat files and write to output
+    with open(fragment_path, "w") as outfile:
+        for i in filenames:
+            for j in i:
+                with open(j, "r") as infile:
+                    for line in infile:
+                        outfile.write(line)
+                os.remove(j)
